import sys
import time
from typing import Dict, Generic, Optional, TypeVar

DEFAULT_CACHE_SIZE = 10 * 1024  # 10KB
DEFAULT_CACHE_TTL = 5  # 5 seconds

T = TypeVar("T")


class CacheProvider(Generic[T]):
    def get(self, key: str) -> Optional[T]:
        pass

    def set(self, key: str, val: T, ttl_override: Optional[int] = None) -> None:
        pass


class CachedItem(Generic[T]):
    def __init__(self, value: T, access_counter: int, size: int, expiration: float):
        self.value = value
        self.access_counter = access_counter
        self.size = size
        self.expiration = expiration


class LocalCache(CacheProvider[T]):
    def __init__(self, max_size: int, ttl: int):
        self.cache: Dict[str, CachedItem[T]] = {}
        self.max_size = max_size
        self.current_size = 0
        self.access_counter = 0
        self.ttl = ttl

    def get(self, key: str) -> Optional[T]:
        if self.max_size == 0:
            return None

        item = self.cache.get(key)
        if item is None:
            return None

        # Check if the item has expired
        if time.time() > item.expiration:
            self.current_size -= item.size
            del self.cache[key]
            return None

        # Update the access counter for LRU eviction
        self.access_counter += 1
        item.access_counter = self.access_counter
        self.cache[key] = item

        return item.value

    def set(self, key: str, val: T, ttl_override: Optional[int] = None) -> None:
        if self.max_size == 0:
            return

        ttl = self.ttl if ttl_override is None else ttl_override
        size = sys.getsizeof(val)

        # Check if the key already exists in the cache
        if key in self.cache:
            item = self.cache[key]
            self.current_size -= item.size
            self.current_size += size
            self.access_counter += 1
            self.cache[key] = CachedItem(val, self.access_counter, size, time.time() + ttl)
            return

        # Evict expired items
        for k, item in list(self.cache.items()):
            if time.time() > item.expiration:
                self.current_size -= item.size
                del self.cache[k]

        # Evict records if the cache size exceeds the max size
        while self.current_size + size > self.max_size:
            oldest_key = None
            oldest_access_counter = float("inf")

            for k, v in self.cache.items():
                if v.access_counter < oldest_access_counter:
                    oldest_key = k
                    oldest_access_counter = v.access_counter

            if oldest_key is not None:
                self.current_size -= self.cache[oldest_key].size
                del self.cache[oldest_key]
            else:
                break

        # Add the new item to the cache
        self.access_counter += 1
        self.cache[key] = CachedItem(val, self.access_counter, size, time.time() + ttl)
        self.current_size += size



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing
import urllib.parse
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError as core_api_error_ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.jsonable_encoder import jsonable_encoder
from ..core.pydantic_utilities import pydantic_v1
from ..core.query_encoder import encode_query
from ..core.remove_none_from_dict import remove_none_from_dict
from ..core.request_options import RequestOptions
from ..errors.bad_request_error import BadRequestError
from ..errors.forbidden_error import ForbiddenError
from ..errors.internal_server_error import InternalServerError
from ..errors.unauthorized_error import UnauthorizedError
from ..types.api_error import ApiError as types_api_error_ApiError
from .types.list_products_response import ListProductsResponse
from .types.upsert_billing_product_response import UpsertBillingProductResponse
from .types.upsert_billing_subscription_response import UpsertBillingSubscriptionResponse

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class BillingClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def upsert_billing_product(
        self,
        *,
        currency: str,
        external_id: str,
        name: str,
        price: float,
        quantity: int,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> UpsertBillingProductResponse:
        """
        Parameters
        ----------
        currency : str

        external_id : str

        name : str

        price : float

        quantity : int

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        UpsertBillingProductResponse
            Created

        Examples
        --------
        from schematic.client import Schematic

        client = Schematic(
            api_key="YOUR_API_KEY",
        )
        client.billing.upsert_billing_product(
            currency="currency",
            external_id="external_id",
            name="name",
            price=1.1,
            quantity=1,
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            method="POST",
            url=urllib.parse.urljoin(f"{self._client_wrapper.get_base_url()}/", "billing/product/upsert"),
            params=encode_query(
                jsonable_encoder(
                    request_options.get("additional_query_parameters") if request_options is not None else None
                )
            ),
            json=jsonable_encoder(
                {"currency": currency, "external_id": external_id, "name": name, "price": price, "quantity": quantity}
            )
            if request_options is None or request_options.get("additional_body_parameters") is None
            else {
                **jsonable_encoder(
                    {
                        "currency": currency,
                        "external_id": external_id,
                        "name": name,
                        "price": price,
                        "quantity": quantity,
                    }
                ),
                **(jsonable_encoder(remove_none_from_dict(request_options.get("additional_body_parameters", {})))),
            },
            headers=jsonable_encoder(
                remove_none_from_dict(
                    {
                        **self._client_wrapper.get_headers(),
                        **(request_options.get("additional_headers", {}) if request_options is not None else {}),
                    }
                )
            ),
            timeout=request_options.get("timeout_in_seconds")
            if request_options is not None and request_options.get("timeout_in_seconds") is not None
            else self._client_wrapper.get_timeout(),
            retries=0,
            max_retries=request_options.get("max_retries") if request_options is not None else 0,  # type: ignore
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(UpsertBillingProductResponse, _response.json())  # type: ignore
        if _response.status_code == 400:
            raise BadRequestError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 401:
            raise UnauthorizedError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        if _response.status_code == 403:
            raise ForbiddenError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 500:
            raise InternalServerError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise core_api_error_ApiError(status_code=_response.status_code, body=_response.text)
        raise core_api_error_ApiError(status_code=_response.status_code, body=_response_json)

    def list_products(
        self,
        *,
        ids: typing.Optional[typing.Union[str, typing.Sequence[str]]] = None,
        name: typing.Optional[str] = None,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ListProductsResponse:
        """
        Parameters
        ----------
        ids : typing.Optional[typing.Union[str, typing.Sequence[str]]]

        name : typing.Optional[str]

        limit : typing.Optional[int]
            Page limit (default 100)

        offset : typing.Optional[int]
            Page offset (default 0)

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ListProductsResponse
            OK

        Examples
        --------
        from schematic.client import Schematic

        client = Schematic(
            api_key="YOUR_API_KEY",
        )
        client.billing.list_products()
        """
        _response = self._client_wrapper.httpx_client.request(
            method="GET",
            url=urllib.parse.urljoin(f"{self._client_wrapper.get_base_url()}/", "billing/products"),
            params=encode_query(
                jsonable_encoder(
                    remove_none_from_dict(
                        {
                            "ids": ids,
                            "name": name,
                            "limit": limit,
                            "offset": offset,
                            **(
                                request_options.get("additional_query_parameters", {})
                                if request_options is not None
                                else {}
                            ),
                        }
                    )
                )
            ),
            headers=jsonable_encoder(
                remove_none_from_dict(
                    {
                        **self._client_wrapper.get_headers(),
                        **(request_options.get("additional_headers", {}) if request_options is not None else {}),
                    }
                )
            ),
            timeout=request_options.get("timeout_in_seconds")
            if request_options is not None and request_options.get("timeout_in_seconds") is not None
            else self._client_wrapper.get_timeout(),
            retries=0,
            max_retries=request_options.get("max_retries") if request_options is not None else 0,  # type: ignore
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(ListProductsResponse, _response.json())  # type: ignore
        if _response.status_code == 400:
            raise BadRequestError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 401:
            raise UnauthorizedError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        if _response.status_code == 403:
            raise ForbiddenError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 500:
            raise InternalServerError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise core_api_error_ApiError(status_code=_response.status_code, body=_response.text)
        raise core_api_error_ApiError(status_code=_response.status_code, body=_response_json)

    def upsert_billing_subscription(
        self,
        *,
        customer_external_id: str,
        expired_at: dt.datetime,
        product_external_ids: typing.Sequence[str],
        subscription_external_id: str,
        interval: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> UpsertBillingSubscriptionResponse:
        """
        Parameters
        ----------
        customer_external_id : str

        expired_at : dt.datetime

        product_external_ids : typing.Sequence[str]

        subscription_external_id : str

        interval : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        UpsertBillingSubscriptionResponse
            Created

        Examples
        --------
        import datetime

        from schematic.client import Schematic

        client = Schematic(
            api_key="YOUR_API_KEY",
        )
        client.billing.upsert_billing_subscription(
            customer_external_id="customer_external_id",
            expired_at=datetime.datetime.fromisoformat(
                "2024-01-15 09:30:00+00:00",
            ),
            product_external_ids=["product_external_ids"],
            subscription_external_id="subscription_external_id",
        )
        """
        _request: typing.Dict[str, typing.Any] = {
            "customer_external_id": customer_external_id,
            "expired_at": expired_at,
            "product_external_ids": product_external_ids,
            "subscription_external_id": subscription_external_id,
        }
        if interval is not OMIT:
            _request["interval"] = interval
        _response = self._client_wrapper.httpx_client.request(
            method="POST",
            url=urllib.parse.urljoin(f"{self._client_wrapper.get_base_url()}/", "billing/subscription/upsert"),
            params=encode_query(
                jsonable_encoder(
                    request_options.get("additional_query_parameters") if request_options is not None else None
                )
            ),
            json=jsonable_encoder(_request)
            if request_options is None or request_options.get("additional_body_parameters") is None
            else {
                **jsonable_encoder(_request),
                **(jsonable_encoder(remove_none_from_dict(request_options.get("additional_body_parameters", {})))),
            },
            headers=jsonable_encoder(
                remove_none_from_dict(
                    {
                        **self._client_wrapper.get_headers(),
                        **(request_options.get("additional_headers", {}) if request_options is not None else {}),
                    }
                )
            ),
            timeout=request_options.get("timeout_in_seconds")
            if request_options is not None and request_options.get("timeout_in_seconds") is not None
            else self._client_wrapper.get_timeout(),
            retries=0,
            max_retries=request_options.get("max_retries") if request_options is not None else 0,  # type: ignore
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(UpsertBillingSubscriptionResponse, _response.json())  # type: ignore
        if _response.status_code == 400:
            raise BadRequestError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 401:
            raise UnauthorizedError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        if _response.status_code == 403:
            raise ForbiddenError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 500:
            raise InternalServerError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise core_api_error_ApiError(status_code=_response.status_code, body=_response.text)
        raise core_api_error_ApiError(status_code=_response.status_code, body=_response_json)


class AsyncBillingClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def upsert_billing_product(
        self,
        *,
        currency: str,
        external_id: str,
        name: str,
        price: float,
        quantity: int,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> UpsertBillingProductResponse:
        """
        Parameters
        ----------
        currency : str

        external_id : str

        name : str

        price : float

        quantity : int

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        UpsertBillingProductResponse
            Created

        Examples
        --------
        from schematic.client import AsyncSchematic

        client = AsyncSchematic(
            api_key="YOUR_API_KEY",
        )
        await client.billing.upsert_billing_product(
            currency="currency",
            external_id="external_id",
            name="name",
            price=1.1,
            quantity=1,
        )
        """
        _response = await self._client_wrapper.httpx_client.request(
            method="POST",
            url=urllib.parse.urljoin(f"{self._client_wrapper.get_base_url()}/", "billing/product/upsert"),
            params=encode_query(
                jsonable_encoder(
                    request_options.get("additional_query_parameters") if request_options is not None else None
                )
            ),
            json=jsonable_encoder(
                {"currency": currency, "external_id": external_id, "name": name, "price": price, "quantity": quantity}
            )
            if request_options is None or request_options.get("additional_body_parameters") is None
            else {
                **jsonable_encoder(
                    {
                        "currency": currency,
                        "external_id": external_id,
                        "name": name,
                        "price": price,
                        "quantity": quantity,
                    }
                ),
                **(jsonable_encoder(remove_none_from_dict(request_options.get("additional_body_parameters", {})))),
            },
            headers=jsonable_encoder(
                remove_none_from_dict(
                    {
                        **self._client_wrapper.get_headers(),
                        **(request_options.get("additional_headers", {}) if request_options is not None else {}),
                    }
                )
            ),
            timeout=request_options.get("timeout_in_seconds")
            if request_options is not None and request_options.get("timeout_in_seconds") is not None
            else self._client_wrapper.get_timeout(),
            retries=0,
            max_retries=request_options.get("max_retries") if request_options is not None else 0,  # type: ignore
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(UpsertBillingProductResponse, _response.json())  # type: ignore
        if _response.status_code == 400:
            raise BadRequestError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 401:
            raise UnauthorizedError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        if _response.status_code == 403:
            raise ForbiddenError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 500:
            raise InternalServerError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise core_api_error_ApiError(status_code=_response.status_code, body=_response.text)
        raise core_api_error_ApiError(status_code=_response.status_code, body=_response_json)

    async def list_products(
        self,
        *,
        ids: typing.Optional[typing.Union[str, typing.Sequence[str]]] = None,
        name: typing.Optional[str] = None,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ListProductsResponse:
        """
        Parameters
        ----------
        ids : typing.Optional[typing.Union[str, typing.Sequence[str]]]

        name : typing.Optional[str]

        limit : typing.Optional[int]
            Page limit (default 100)

        offset : typing.Optional[int]
            Page offset (default 0)

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ListProductsResponse
            OK

        Examples
        --------
        from schematic.client import AsyncSchematic

        client = AsyncSchematic(
            api_key="YOUR_API_KEY",
        )
        await client.billing.list_products()
        """
        _response = await self._client_wrapper.httpx_client.request(
            method="GET",
            url=urllib.parse.urljoin(f"{self._client_wrapper.get_base_url()}/", "billing/products"),
            params=encode_query(
                jsonable_encoder(
                    remove_none_from_dict(
                        {
                            "ids": ids,
                            "name": name,
                            "limit": limit,
                            "offset": offset,
                            **(
                                request_options.get("additional_query_parameters", {})
                                if request_options is not None
                                else {}
                            ),
                        }
                    )
                )
            ),
            headers=jsonable_encoder(
                remove_none_from_dict(
                    {
                        **self._client_wrapper.get_headers(),
                        **(request_options.get("additional_headers", {}) if request_options is not None else {}),
                    }
                )
            ),
            timeout=request_options.get("timeout_in_seconds")
            if request_options is not None and request_options.get("timeout_in_seconds") is not None
            else self._client_wrapper.get_timeout(),
            retries=0,
            max_retries=request_options.get("max_retries") if request_options is not None else 0,  # type: ignore
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(ListProductsResponse, _response.json())  # type: ignore
        if _response.status_code == 400:
            raise BadRequestError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 401:
            raise UnauthorizedError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        if _response.status_code == 403:
            raise ForbiddenError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 500:
            raise InternalServerError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise core_api_error_ApiError(status_code=_response.status_code, body=_response.text)
        raise core_api_error_ApiError(status_code=_response.status_code, body=_response_json)

    async def upsert_billing_subscription(
        self,
        *,
        customer_external_id: str,
        expired_at: dt.datetime,
        product_external_ids: typing.Sequence[str],
        subscription_external_id: str,
        interval: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> UpsertBillingSubscriptionResponse:
        """
        Parameters
        ----------
        customer_external_id : str

        expired_at : dt.datetime

        product_external_ids : typing.Sequence[str]

        subscription_external_id : str

        interval : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        UpsertBillingSubscriptionResponse
            Created

        Examples
        --------
        import datetime

        from schematic.client import AsyncSchematic

        client = AsyncSchematic(
            api_key="YOUR_API_KEY",
        )
        await client.billing.upsert_billing_subscription(
            customer_external_id="customer_external_id",
            expired_at=datetime.datetime.fromisoformat(
                "2024-01-15 09:30:00+00:00",
            ),
            product_external_ids=["product_external_ids"],
            subscription_external_id="subscription_external_id",
        )
        """
        _request: typing.Dict[str, typing.Any] = {
            "customer_external_id": customer_external_id,
            "expired_at": expired_at,
            "product_external_ids": product_external_ids,
            "subscription_external_id": subscription_external_id,
        }
        if interval is not OMIT:
            _request["interval"] = interval
        _response = await self._client_wrapper.httpx_client.request(
            method="POST",
            url=urllib.parse.urljoin(f"{self._client_wrapper.get_base_url()}/", "billing/subscription/upsert"),
            params=encode_query(
                jsonable_encoder(
                    request_options.get("additional_query_parameters") if request_options is not None else None
                )
            ),
            json=jsonable_encoder(_request)
            if request_options is None or request_options.get("additional_body_parameters") is None
            else {
                **jsonable_encoder(_request),
                **(jsonable_encoder(remove_none_from_dict(request_options.get("additional_body_parameters", {})))),
            },
            headers=jsonable_encoder(
                remove_none_from_dict(
                    {
                        **self._client_wrapper.get_headers(),
                        **(request_options.get("additional_headers", {}) if request_options is not None else {}),
                    }
                )
            ),
            timeout=request_options.get("timeout_in_seconds")
            if request_options is not None and request_options.get("timeout_in_seconds") is not None
            else self._client_wrapper.get_timeout(),
            retries=0,
            max_retries=request_options.get("max_retries") if request_options is not None else 0,  # type: ignore
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(UpsertBillingSubscriptionResponse, _response.json())  # type: ignore
        if _response.status_code == 400:
            raise BadRequestError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 401:
            raise UnauthorizedError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        if _response.status_code == 403:
            raise ForbiddenError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 500:
            raise InternalServerError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise core_api_error_ApiError(status_code=_response.status_code, body=_response.text)
        raise core_api_error_ApiError(status_code=_response.status_code, body=_response_json)



# This file was auto-generated by Fern from our API Definition.

import typing
import urllib.parse
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError as core_api_error_ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.jsonable_encoder import jsonable_encoder
from ..core.pydantic_utilities import pydantic_v1
from ..core.query_encoder import encode_query
from ..core.remove_none_from_dict import remove_none_from_dict
from ..core.request_options import RequestOptions
from ..errors.bad_request_error import BadRequestError
from ..errors.forbidden_error import ForbiddenError
from ..errors.internal_server_error import InternalServerError
from ..errors.unauthorized_error import UnauthorizedError
from ..types.api_error import ApiError as types_api_error_ApiError
from .types.list_crm_products_response import ListCrmProductsResponse
from .types.upsert_crm_deal_response import UpsertCrmDealResponse
from .types.upsert_crm_product_response import UpsertCrmProductResponse
from .types.upsert_deal_line_item_association_response import UpsertDealLineItemAssociationResponse
from .types.upsert_line_item_response import UpsertLineItemResponse

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class CrmClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def upsert_deal_line_item_association(
        self,
        *,
        deal_external_id: str,
        line_item_external_id: str,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> UpsertDealLineItemAssociationResponse:
        """
        Parameters
        ----------
        deal_external_id : str

        line_item_external_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        UpsertDealLineItemAssociationResponse
            Created

        Examples
        --------
        from schematic.client import Schematic

        client = Schematic(
            api_key="YOUR_API_KEY",
        )
        client.crm.upsert_deal_line_item_association(
            deal_external_id="deal_external_id",
            line_item_external_id="line_item_external_id",
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            method="POST",
            url=urllib.parse.urljoin(f"{self._client_wrapper.get_base_url()}/", "crm/associations/deal-line-item"),
            params=encode_query(
                jsonable_encoder(
                    request_options.get("additional_query_parameters") if request_options is not None else None
                )
            ),
            json=jsonable_encoder(
                {"deal_external_id": deal_external_id, "line_item_external_id": line_item_external_id}
            )
            if request_options is None or request_options.get("additional_body_parameters") is None
            else {
                **jsonable_encoder(
                    {"deal_external_id": deal_external_id, "line_item_external_id": line_item_external_id}
                ),
                **(jsonable_encoder(remove_none_from_dict(request_options.get("additional_body_parameters", {})))),
            },
            headers=jsonable_encoder(
                remove_none_from_dict(
                    {
                        **self._client_wrapper.get_headers(),
                        **(request_options.get("additional_headers", {}) if request_options is not None else {}),
                    }
                )
            ),
            timeout=request_options.get("timeout_in_seconds")
            if request_options is not None and request_options.get("timeout_in_seconds") is not None
            else self._client_wrapper.get_timeout(),
            retries=0,
            max_retries=request_options.get("max_retries") if request_options is not None else 0,  # type: ignore
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(UpsertDealLineItemAssociationResponse, _response.json())  # type: ignore
        if _response.status_code == 400:
            raise BadRequestError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 401:
            raise UnauthorizedError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        if _response.status_code == 403:
            raise ForbiddenError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 500:
            raise InternalServerError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise core_api_error_ApiError(status_code=_response.status_code, body=_response.text)
        raise core_api_error_ApiError(status_code=_response.status_code, body=_response_json)

    def upsert_line_item(
        self,
        *,
        amount: str,
        interval: str,
        line_item_external_id: str,
        product_external_id: str,
        quantity: int,
        term_month: typing.Optional[int] = OMIT,
        discount_percentage: typing.Optional[str] = OMIT,
        total_discount: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> UpsertLineItemResponse:
        """
        Parameters
        ----------
        amount : str

        interval : str

        line_item_external_id : str

        product_external_id : str

        quantity : int

        term_month : typing.Optional[int]

        discount_percentage : typing.Optional[str]

        total_discount : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        UpsertLineItemResponse
            Created

        Examples
        --------
        from schematic.client import Schematic

        client = Schematic(
            api_key="YOUR_API_KEY",
        )
        client.crm.upsert_line_item(
            amount="amount",
            interval="interval",
            line_item_external_id="line_item_external_id",
            product_external_id="product_external_id",
            quantity=1,
        )
        """
        _request: typing.Dict[str, typing.Any] = {
            "amount": amount,
            "interval": interval,
            "line_item_external_id": line_item_external_id,
            "product_external_id": product_external_id,
            "quantity": quantity,
        }
        if term_month is not OMIT:
            _request["TermMonth"] = term_month
        if discount_percentage is not OMIT:
            _request["discount_percentage"] = discount_percentage
        if total_discount is not OMIT:
            _request["total_discount"] = total_discount
        _response = self._client_wrapper.httpx_client.request(
            method="POST",
            url=urllib.parse.urljoin(f"{self._client_wrapper.get_base_url()}/", "crm/deal-line-item/upsert"),
            params=encode_query(
                jsonable_encoder(
                    request_options.get("additional_query_parameters") if request_options is not None else None
                )
            ),
            json=jsonable_encoder(_request)
            if request_options is None or request_options.get("additional_body_parameters") is None
            else {
                **jsonable_encoder(_request),
                **(jsonable_encoder(remove_none_from_dict(request_options.get("additional_body_parameters", {})))),
            },
            headers=jsonable_encoder(
                remove_none_from_dict(
                    {
                        **self._client_wrapper.get_headers(),
                        **(request_options.get("additional_headers", {}) if request_options is not None else {}),
                    }
                )
            ),
            timeout=request_options.get("timeout_in_seconds")
            if request_options is not None and request_options.get("timeout_in_seconds") is not None
            else self._client_wrapper.get_timeout(),
            retries=0,
            max_retries=request_options.get("max_retries") if request_options is not None else 0,  # type: ignore
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(UpsertLineItemResponse, _response.json())  # type: ignore
        if _response.status_code == 400:
            raise BadRequestError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 401:
            raise UnauthorizedError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        if _response.status_code == 403:
            raise ForbiddenError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 500:
            raise InternalServerError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise core_api_error_ApiError(status_code=_response.status_code, body=_response.text)
        raise core_api_error_ApiError(status_code=_response.status_code, body=_response_json)

    def upsert_crm_deal(
        self,
        *,
        crm_company_key: str,
        crm_type: str,
        deal_external_id: str,
        arr: typing.Optional[str] = OMIT,
        crm_company_id: typing.Optional[str] = OMIT,
        crm_product_id: typing.Optional[str] = OMIT,
        deal_name: typing.Optional[str] = OMIT,
        deal_stage: typing.Optional[str] = OMIT,
        mrr: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> UpsertCrmDealResponse:
        """
        Parameters
        ----------
        crm_company_key : str

        crm_type : str

        deal_external_id : str

        arr : typing.Optional[str]

        crm_company_id : typing.Optional[str]

        crm_product_id : typing.Optional[str]

        deal_name : typing.Optional[str]

        deal_stage : typing.Optional[str]

        mrr : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        UpsertCrmDealResponse
            Created

        Examples
        --------
        from schematic.client import Schematic

        client = Schematic(
            api_key="YOUR_API_KEY",
        )
        client.crm.upsert_crm_deal(
            crm_company_key="crm_company_key",
            crm_type="crm_type",
            deal_external_id="deal_external_id",
        )
        """
        _request: typing.Dict[str, typing.Any] = {
            "crm_company_key": crm_company_key,
            "crm_type": crm_type,
            "deal_external_id": deal_external_id,
        }
        if arr is not OMIT:
            _request["arr"] = arr
        if crm_company_id is not OMIT:
            _request["crm_company_id"] = crm_company_id
        if crm_product_id is not OMIT:
            _request["crm_product_id"] = crm_product_id
        if deal_name is not OMIT:
            _request["deal_name"] = deal_name
        if deal_stage is not OMIT:
            _request["deal_stage"] = deal_stage
        if mrr is not OMIT:
            _request["mrr"] = mrr
        _response = self._client_wrapper.httpx_client.request(
            method="POST",
            url=urllib.parse.urljoin(f"{self._client_wrapper.get_base_url()}/", "crm/deals/upsert"),
            params=encode_query(
                jsonable_encoder(
                    request_options.get("additional_query_parameters") if request_options is not None else None
                )
            ),
            json=jsonable_encoder(_request)
            if request_options is None or request_options.get("additional_body_parameters") is None
            else {
                **jsonable_encoder(_request),
                **(jsonable_encoder(remove_none_from_dict(request_options.get("additional_body_parameters", {})))),
            },
            headers=jsonable_encoder(
                remove_none_from_dict(
                    {
                        **self._client_wrapper.get_headers(),
                        **(request_options.get("additional_headers", {}) if request_options is not None else {}),
                    }
                )
            ),
            timeout=request_options.get("timeout_in_seconds")
            if request_options is not None and request_options.get("timeout_in_seconds") is not None
            else self._client_wrapper.get_timeout(),
            retries=0,
            max_retries=request_options.get("max_retries") if request_options is not None else 0,  # type: ignore
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(UpsertCrmDealResponse, _response.json())  # type: ignore
        if _response.status_code == 400:
            raise BadRequestError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 401:
            raise UnauthorizedError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        if _response.status_code == 403:
            raise ForbiddenError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 500:
            raise InternalServerError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise core_api_error_ApiError(status_code=_response.status_code, body=_response.text)
        raise core_api_error_ApiError(status_code=_response.status_code, body=_response_json)

    def list_crm_products(
        self,
        *,
        ids: typing.Optional[typing.Union[str, typing.Sequence[str]]] = None,
        name: typing.Optional[str] = None,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ListCrmProductsResponse:
        """
        Parameters
        ----------
        ids : typing.Optional[typing.Union[str, typing.Sequence[str]]]

        name : typing.Optional[str]

        limit : typing.Optional[int]
            Page limit (default 100)

        offset : typing.Optional[int]
            Page offset (default 0)

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ListCrmProductsResponse
            OK

        Examples
        --------
        from schematic.client import Schematic

        client = Schematic(
            api_key="YOUR_API_KEY",
        )
        client.crm.list_crm_products()
        """
        _response = self._client_wrapper.httpx_client.request(
            method="GET",
            url=urllib.parse.urljoin(f"{self._client_wrapper.get_base_url()}/", "crm/products"),
            params=encode_query(
                jsonable_encoder(
                    remove_none_from_dict(
                        {
                            "ids": ids,
                            "name": name,
                            "limit": limit,
                            "offset": offset,
                            **(
                                request_options.get("additional_query_parameters", {})
                                if request_options is not None
                                else {}
                            ),
                        }
                    )
                )
            ),
            headers=jsonable_encoder(
                remove_none_from_dict(
                    {
                        **self._client_wrapper.get_headers(),
                        **(request_options.get("additional_headers", {}) if request_options is not None else {}),
                    }
                )
            ),
            timeout=request_options.get("timeout_in_seconds")
            if request_options is not None and request_options.get("timeout_in_seconds") is not None
            else self._client_wrapper.get_timeout(),
            retries=0,
            max_retries=request_options.get("max_retries") if request_options is not None else 0,  # type: ignore
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(ListCrmProductsResponse, _response.json())  # type: ignore
        if _response.status_code == 400:
            raise BadRequestError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 401:
            raise UnauthorizedError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        if _response.status_code == 403:
            raise ForbiddenError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 500:
            raise InternalServerError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise core_api_error_ApiError(status_code=_response.status_code, body=_response.text)
        raise core_api_error_ApiError(status_code=_response.status_code, body=_response_json)

    def upsert_crm_product(
        self,
        *,
        currency: str,
        description: str,
        external_id: str,
        interval: str,
        name: str,
        price: str,
        quantity: int,
        sku: str,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> UpsertCrmProductResponse:
        """
        Parameters
        ----------
        currency : str

        description : str

        external_id : str

        interval : str

        name : str

        price : str

        quantity : int

        sku : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        UpsertCrmProductResponse
            Created

        Examples
        --------
        from schematic.client import Schematic

        client = Schematic(
            api_key="YOUR_API_KEY",
        )
        client.crm.upsert_crm_product(
            currency="currency",
            description="description",
            external_id="external_id",
            interval="interval",
            name="name",
            price="price",
            quantity=1,
            sku="sku",
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            method="POST",
            url=urllib.parse.urljoin(f"{self._client_wrapper.get_base_url()}/", "crm/products/upsert"),
            params=encode_query(
                jsonable_encoder(
                    request_options.get("additional_query_parameters") if request_options is not None else None
                )
            ),
            json=jsonable_encoder(
                {
                    "currency": currency,
                    "description": description,
                    "external_id": external_id,
                    "interval": interval,
                    "name": name,
                    "price": price,
                    "quantity": quantity,
                    "sku": sku,
                }
            )
            if request_options is None or request_options.get("additional_body_parameters") is None
            else {
                **jsonable_encoder(
                    {
                        "currency": currency,
                        "description": description,
                        "external_id": external_id,
                        "interval": interval,
                        "name": name,
                        "price": price,
                        "quantity": quantity,
                        "sku": sku,
                    }
                ),
                **(jsonable_encoder(remove_none_from_dict(request_options.get("additional_body_parameters", {})))),
            },
            headers=jsonable_encoder(
                remove_none_from_dict(
                    {
                        **self._client_wrapper.get_headers(),
                        **(request_options.get("additional_headers", {}) if request_options is not None else {}),
                    }
                )
            ),
            timeout=request_options.get("timeout_in_seconds")
            if request_options is not None and request_options.get("timeout_in_seconds") is not None
            else self._client_wrapper.get_timeout(),
            retries=0,
            max_retries=request_options.get("max_retries") if request_options is not None else 0,  # type: ignore
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(UpsertCrmProductResponse, _response.json())  # type: ignore
        if _response.status_code == 400:
            raise BadRequestError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 401:
            raise UnauthorizedError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        if _response.status_code == 403:
            raise ForbiddenError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 500:
            raise InternalServerError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise core_api_error_ApiError(status_code=_response.status_code, body=_response.text)
        raise core_api_error_ApiError(status_code=_response.status_code, body=_response_json)


class AsyncCrmClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def upsert_deal_line_item_association(
        self,
        *,
        deal_external_id: str,
        line_item_external_id: str,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> UpsertDealLineItemAssociationResponse:
        """
        Parameters
        ----------
        deal_external_id : str

        line_item_external_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        UpsertDealLineItemAssociationResponse
            Created

        Examples
        --------
        from schematic.client import AsyncSchematic

        client = AsyncSchematic(
            api_key="YOUR_API_KEY",
        )
        await client.crm.upsert_deal_line_item_association(
            deal_external_id="deal_external_id",
            line_item_external_id="line_item_external_id",
        )
        """
        _response = await self._client_wrapper.httpx_client.request(
            method="POST",
            url=urllib.parse.urljoin(f"{self._client_wrapper.get_base_url()}/", "crm/associations/deal-line-item"),
            params=encode_query(
                jsonable_encoder(
                    request_options.get("additional_query_parameters") if request_options is not None else None
                )
            ),
            json=jsonable_encoder(
                {"deal_external_id": deal_external_id, "line_item_external_id": line_item_external_id}
            )
            if request_options is None or request_options.get("additional_body_parameters") is None
            else {
                **jsonable_encoder(
                    {"deal_external_id": deal_external_id, "line_item_external_id": line_item_external_id}
                ),
                **(jsonable_encoder(remove_none_from_dict(request_options.get("additional_body_parameters", {})))),
            },
            headers=jsonable_encoder(
                remove_none_from_dict(
                    {
                        **self._client_wrapper.get_headers(),
                        **(request_options.get("additional_headers", {}) if request_options is not None else {}),
                    }
                )
            ),
            timeout=request_options.get("timeout_in_seconds")
            if request_options is not None and request_options.get("timeout_in_seconds") is not None
            else self._client_wrapper.get_timeout(),
            retries=0,
            max_retries=request_options.get("max_retries") if request_options is not None else 0,  # type: ignore
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(UpsertDealLineItemAssociationResponse, _response.json())  # type: ignore
        if _response.status_code == 400:
            raise BadRequestError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 401:
            raise UnauthorizedError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        if _response.status_code == 403:
            raise ForbiddenError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 500:
            raise InternalServerError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise core_api_error_ApiError(status_code=_response.status_code, body=_response.text)
        raise core_api_error_ApiError(status_code=_response.status_code, body=_response_json)

    async def upsert_line_item(
        self,
        *,
        amount: str,
        interval: str,
        line_item_external_id: str,
        product_external_id: str,
        quantity: int,
        term_month: typing.Optional[int] = OMIT,
        discount_percentage: typing.Optional[str] = OMIT,
        total_discount: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> UpsertLineItemResponse:
        """
        Parameters
        ----------
        amount : str

        interval : str

        line_item_external_id : str

        product_external_id : str

        quantity : int

        term_month : typing.Optional[int]

        discount_percentage : typing.Optional[str]

        total_discount : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        UpsertLineItemResponse
            Created

        Examples
        --------
        from schematic.client import AsyncSchematic

        client = AsyncSchematic(
            api_key="YOUR_API_KEY",
        )
        await client.crm.upsert_line_item(
            amount="amount",
            interval="interval",
            line_item_external_id="line_item_external_id",
            product_external_id="product_external_id",
            quantity=1,
        )
        """
        _request: typing.Dict[str, typing.Any] = {
            "amount": amount,
            "interval": interval,
            "line_item_external_id": line_item_external_id,
            "product_external_id": product_external_id,
            "quantity": quantity,
        }
        if term_month is not OMIT:
            _request["TermMonth"] = term_month
        if discount_percentage is not OMIT:
            _request["discount_percentage"] = discount_percentage
        if total_discount is not OMIT:
            _request["total_discount"] = total_discount
        _response = await self._client_wrapper.httpx_client.request(
            method="POST",
            url=urllib.parse.urljoin(f"{self._client_wrapper.get_base_url()}/", "crm/deal-line-item/upsert"),
            params=encode_query(
                jsonable_encoder(
                    request_options.get("additional_query_parameters") if request_options is not None else None
                )
            ),
            json=jsonable_encoder(_request)
            if request_options is None or request_options.get("additional_body_parameters") is None
            else {
                **jsonable_encoder(_request),
                **(jsonable_encoder(remove_none_from_dict(request_options.get("additional_body_parameters", {})))),
            },
            headers=jsonable_encoder(
                remove_none_from_dict(
                    {
                        **self._client_wrapper.get_headers(),
                        **(request_options.get("additional_headers", {}) if request_options is not None else {}),
                    }
                )
            ),
            timeout=request_options.get("timeout_in_seconds")
            if request_options is not None and request_options.get("timeout_in_seconds") is not None
            else self._client_wrapper.get_timeout(),
            retries=0,
            max_retries=request_options.get("max_retries") if request_options is not None else 0,  # type: ignore
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(UpsertLineItemResponse, _response.json())  # type: ignore
        if _response.status_code == 400:
            raise BadRequestError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 401:
            raise UnauthorizedError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        if _response.status_code == 403:
            raise ForbiddenError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 500:
            raise InternalServerError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise core_api_error_ApiError(status_code=_response.status_code, body=_response.text)
        raise core_api_error_ApiError(status_code=_response.status_code, body=_response_json)

    async def upsert_crm_deal(
        self,
        *,
        crm_company_key: str,
        crm_type: str,
        deal_external_id: str,
        arr: typing.Optional[str] = OMIT,
        crm_company_id: typing.Optional[str] = OMIT,
        crm_product_id: typing.Optional[str] = OMIT,
        deal_name: typing.Optional[str] = OMIT,
        deal_stage: typing.Optional[str] = OMIT,
        mrr: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> UpsertCrmDealResponse:
        """
        Parameters
        ----------
        crm_company_key : str

        crm_type : str

        deal_external_id : str

        arr : typing.Optional[str]

        crm_company_id : typing.Optional[str]

        crm_product_id : typing.Optional[str]

        deal_name : typing.Optional[str]

        deal_stage : typing.Optional[str]

        mrr : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        UpsertCrmDealResponse
            Created

        Examples
        --------
        from schematic.client import AsyncSchematic

        client = AsyncSchematic(
            api_key="YOUR_API_KEY",
        )
        await client.crm.upsert_crm_deal(
            crm_company_key="crm_company_key",
            crm_type="crm_type",
            deal_external_id="deal_external_id",
        )
        """
        _request: typing.Dict[str, typing.Any] = {
            "crm_company_key": crm_company_key,
            "crm_type": crm_type,
            "deal_external_id": deal_external_id,
        }
        if arr is not OMIT:
            _request["arr"] = arr
        if crm_company_id is not OMIT:
            _request["crm_company_id"] = crm_company_id
        if crm_product_id is not OMIT:
            _request["crm_product_id"] = crm_product_id
        if deal_name is not OMIT:
            _request["deal_name"] = deal_name
        if deal_stage is not OMIT:
            _request["deal_stage"] = deal_stage
        if mrr is not OMIT:
            _request["mrr"] = mrr
        _response = await self._client_wrapper.httpx_client.request(
            method="POST",
            url=urllib.parse.urljoin(f"{self._client_wrapper.get_base_url()}/", "crm/deals/upsert"),
            params=encode_query(
                jsonable_encoder(
                    request_options.get("additional_query_parameters") if request_options is not None else None
                )
            ),
            json=jsonable_encoder(_request)
            if request_options is None or request_options.get("additional_body_parameters") is None
            else {
                **jsonable_encoder(_request),
                **(jsonable_encoder(remove_none_from_dict(request_options.get("additional_body_parameters", {})))),
            },
            headers=jsonable_encoder(
                remove_none_from_dict(
                    {
                        **self._client_wrapper.get_headers(),
                        **(request_options.get("additional_headers", {}) if request_options is not None else {}),
                    }
                )
            ),
            timeout=request_options.get("timeout_in_seconds")
            if request_options is not None and request_options.get("timeout_in_seconds") is not None
            else self._client_wrapper.get_timeout(),
            retries=0,
            max_retries=request_options.get("max_retries") if request_options is not None else 0,  # type: ignore
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(UpsertCrmDealResponse, _response.json())  # type: ignore
        if _response.status_code == 400:
            raise BadRequestError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 401:
            raise UnauthorizedError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        if _response.status_code == 403:
            raise ForbiddenError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 500:
            raise InternalServerError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise core_api_error_ApiError(status_code=_response.status_code, body=_response.text)
        raise core_api_error_ApiError(status_code=_response.status_code, body=_response_json)

    async def list_crm_products(
        self,
        *,
        ids: typing.Optional[typing.Union[str, typing.Sequence[str]]] = None,
        name: typing.Optional[str] = None,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ListCrmProductsResponse:
        """
        Parameters
        ----------
        ids : typing.Optional[typing.Union[str, typing.Sequence[str]]]

        name : typing.Optional[str]

        limit : typing.Optional[int]
            Page limit (default 100)

        offset : typing.Optional[int]
            Page offset (default 0)

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ListCrmProductsResponse
            OK

        Examples
        --------
        from schematic.client import AsyncSchematic

        client = AsyncSchematic(
            api_key="YOUR_API_KEY",
        )
        await client.crm.list_crm_products()
        """
        _response = await self._client_wrapper.httpx_client.request(
            method="GET",
            url=urllib.parse.urljoin(f"{self._client_wrapper.get_base_url()}/", "crm/products"),
            params=encode_query(
                jsonable_encoder(
                    remove_none_from_dict(
                        {
                            "ids": ids,
                            "name": name,
                            "limit": limit,
                            "offset": offset,
                            **(
                                request_options.get("additional_query_parameters", {})
                                if request_options is not None
                                else {}
                            ),
                        }
                    )
                )
            ),
            headers=jsonable_encoder(
                remove_none_from_dict(
                    {
                        **self._client_wrapper.get_headers(),
                        **(request_options.get("additional_headers", {}) if request_options is not None else {}),
                    }
                )
            ),
            timeout=request_options.get("timeout_in_seconds")
            if request_options is not None and request_options.get("timeout_in_seconds") is not None
            else self._client_wrapper.get_timeout(),
            retries=0,
            max_retries=request_options.get("max_retries") if request_options is not None else 0,  # type: ignore
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(ListCrmProductsResponse, _response.json())  # type: ignore
        if _response.status_code == 400:
            raise BadRequestError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 401:
            raise UnauthorizedError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        if _response.status_code == 403:
            raise ForbiddenError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 500:
            raise InternalServerError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise core_api_error_ApiError(status_code=_response.status_code, body=_response.text)
        raise core_api_error_ApiError(status_code=_response.status_code, body=_response_json)

    async def upsert_crm_product(
        self,
        *,
        currency: str,
        description: str,
        external_id: str,
        interval: str,
        name: str,
        price: str,
        quantity: int,
        sku: str,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> UpsertCrmProductResponse:
        """
        Parameters
        ----------
        currency : str

        description : str

        external_id : str

        interval : str

        name : str

        price : str

        quantity : int

        sku : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        UpsertCrmProductResponse
            Created

        Examples
        --------
        from schematic.client import AsyncSchematic

        client = AsyncSchematic(
            api_key="YOUR_API_KEY",
        )
        await client.crm.upsert_crm_product(
            currency="currency",
            description="description",
            external_id="external_id",
            interval="interval",
            name="name",
            price="price",
            quantity=1,
            sku="sku",
        )
        """
        _response = await self._client_wrapper.httpx_client.request(
            method="POST",
            url=urllib.parse.urljoin(f"{self._client_wrapper.get_base_url()}/", "crm/products/upsert"),
            params=encode_query(
                jsonable_encoder(
                    request_options.get("additional_query_parameters") if request_options is not None else None
                )
            ),
            json=jsonable_encoder(
                {
                    "currency": currency,
                    "description": description,
                    "external_id": external_id,
                    "interval": interval,
                    "name": name,
                    "price": price,
                    "quantity": quantity,
                    "sku": sku,
                }
            )
            if request_options is None or request_options.get("additional_body_parameters") is None
            else {
                **jsonable_encoder(
                    {
                        "currency": currency,
                        "description": description,
                        "external_id": external_id,
                        "interval": interval,
                        "name": name,
                        "price": price,
                        "quantity": quantity,
                        "sku": sku,
                    }
                ),
                **(jsonable_encoder(remove_none_from_dict(request_options.get("additional_body_parameters", {})))),
            },
            headers=jsonable_encoder(
                remove_none_from_dict(
                    {
                        **self._client_wrapper.get_headers(),
                        **(request_options.get("additional_headers", {}) if request_options is not None else {}),
                    }
                )
            ),
            timeout=request_options.get("timeout_in_seconds")
            if request_options is not None and request_options.get("timeout_in_seconds") is not None
            else self._client_wrapper.get_timeout(),
            retries=0,
            max_retries=request_options.get("max_retries") if request_options is not None else 0,  # type: ignore
        )
        if 200 <= _response.status_code < 300:
            return pydantic_v1.parse_obj_as(UpsertCrmProductResponse, _response.json())  # type: ignore
        if _response.status_code == 400:
            raise BadRequestError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 401:
            raise UnauthorizedError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        if _response.status_code == 403:
            raise ForbiddenError(pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json()))  # type: ignore
        if _response.status_code == 500:
            raise InternalServerError(
                pydantic_v1.parse_obj_as(types_api_error_ApiError, _response.json())  # type: ignore
            )
        try:
            _response_json = _response.json()
        except JSONDecodeError:
            raise core_api_error_ApiError(status_code=_response.status_code, body=_response.text)
        raise core_api_error_ApiError(status_code=_response.status_code, body=_response_json)



import asyncio
import atexit
import logging
import signal
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import httpx

from .base_client import AsyncBaseSchematic, BaseSchematic
from .cache import DEFAULT_CACHE_SIZE, DEFAULT_CACHE_TTL, CacheProvider, LocalCache
from .event_buffer import AsyncEventBuffer, EventBuffer
from .http_client import AsyncOfflineHTTPClient, OfflineHTTPClient
from .logging import get_default_logger
from .types import (
    CreateEventRequestBody,
    EventBody,
    EventBodyIdentify,
    EventBodyIdentifyCompany,
    EventBodyTrack,
)


@dataclass
class SchematicConfig:
    base_url: Optional[str] = None
    event_buffer_period: Optional[int] = None
    flag_defaults: Optional[Dict[str, bool]] = None
    follow_redirects: Optional[bool] = True
    httpx_client: Optional[httpx.Client] = None
    logger: Optional[logging.Logger] = None
    offline: bool = False
    timeout: Optional[float] = None
    cache_providers: Optional[List[CacheProvider[bool]]] = None


class Schematic(BaseSchematic):
    def __init__(self, api_key: str, config: Optional[SchematicConfig] = None):
        config = config or SchematicConfig()
        httpx_client = OfflineHTTPClient() if config.offline else config.httpx_client

        super().__init__(
            api_key=api_key,
            base_url=config.base_url,
            follow_redirects=config.follow_redirects,
            httpx_client=httpx_client,
            timeout=config.timeout,
        )
        self.event_buffer_period = config.event_buffer_period
        self.logger = config.logger or get_default_logger()
        self.flag_defaults = config.flag_defaults or {}
        self.event_buffer = EventBuffer(
            events_api=self.events,
            logger=self.logger,
            period=self.event_buffer_period,
        )
        self.flag_check_cache_providers = config.cache_providers or [
            LocalCache[bool](DEFAULT_CACHE_SIZE, DEFAULT_CACHE_TTL)
        ]
        self.offline = config.offline

        atexit.register(self.shutdown)

    def initialize(self) -> None:
        pass

    def shutdown(self) -> None:
        self.event_buffer.stop()

    def check_flag(
        self,
        flag_key: str,
        company: Optional[Dict[str, str]] = None,
        user: Optional[Dict[str, str]] = None,
    ) -> bool:
        if self.offline:
            return self._get_flag_default(flag_key)

        try:
            cache_key = (
                flag_key + ":" + str(company) + ":" + str(user) if (company or user) else flag_key
            )

            for provider in self.flag_check_cache_providers:
                cached_value = provider.get(cache_key)
                if cached_value is not None:
                    return cached_value

            resp = self.features.check_flag(flag_key, company=company, user=user)
            if resp is None:
                return self._get_flag_default(flag_key)

            for provider in self.flag_check_cache_providers:
                provider.set(cache_key, resp.data.value)

            return resp.data.value
        except Exception as e:
            self.logger.error(e)
            return self._get_flag_default(flag_key)

    def identify(
        self,
        keys: Dict[str, str],
        company: Optional[EventBodyIdentifyCompany] = None,
        name: Optional[str] = None,
        traits: Optional[Dict[str, Any]] = None,
    ) -> None:
        self._enqueue_event(
            "identify",
            EventBodyIdentify(
                company=company,
                keys=keys,
                name=name,
                traits=traits,
            ),
        )

    def track(
        self,
        event: str,
        company: Optional[Dict[str, str]] = None,
        user: Optional[Dict[str, str]] = None,
        traits: Optional[Dict[str, Any]] = None,
    ) -> None:
        self._enqueue_event(
            "track",
            EventBodyTrack(
                company=company,
                event=event,
                traits=traits,
                user=user,
            ),
        )

    def _enqueue_event(self, event_type: str, body: EventBody) -> None:
        if self.offline:
            return
        try:
            event_body = CreateEventRequestBody(event_type=event_type, body=body)
            self.event_buffer.push(event_body)
        except Exception as e:
            self.logger.error(e)

    def _get_flag_default(self, flag_key: str) -> bool:
        return self.flag_defaults.get(flag_key, False)


@dataclass
class AsyncSchematicConfig:
    base_url: Optional[str] = None
    event_buffer_period: Optional[int] = None
    flag_defaults: Optional[Dict[str, bool]] = None
    follow_redirects: Optional[bool] = True
    httpx_client: Optional[httpx.AsyncClient] = None
    logger: Optional[logging.Logger] = None
    offline: bool = False
    timeout: Optional[float] = None
    cache_providers: Optional[List[CacheProvider[bool]]] = None


class AsyncSchematic(AsyncBaseSchematic):
    def __init__(self, api_key: str, config: Optional[AsyncSchematicConfig] = None):
        config = config or AsyncSchematicConfig()
        httpx_client = AsyncOfflineHTTPClient() if config.offline else config.httpx_client

        super().__init__(
            api_key=api_key,
            base_url=config.base_url,
            follow_redirects=config.follow_redirects,
            httpx_client=httpx_client,
            timeout=config.timeout,
        )
        self.event_buffer_period = config.event_buffer_period
        self.logger = config.logger or get_default_logger()
        self.flag_defaults = config.flag_defaults or {}
        self.event_buffer = AsyncEventBuffer(
            events_api=self.events,
            logger=self.logger,
            period=self.event_buffer_period,
        )
        self.flag_check_cache_providers = config.cache_providers or [
            LocalCache[bool](DEFAULT_CACHE_SIZE, DEFAULT_CACHE_TTL)
        ]
        self.offline = config.offline
        for sig in (signal.SIGINT, signal.SIGTERM):
            signal.signal(sig, self._shutdown_handler)

    async def initialize(self) -> None:
        pass

    async def check_flag(
        self,
        flag_key: str,
        company: Optional[Dict[str, str]] = None,
        user: Optional[Dict[str, str]] = None,
    ) -> bool:
        if self.offline:
            return self._get_flag_default(flag_key)

        try:
            cache_key = (
                flag_key + ":" + str(company) + ":" + str(user) if (company or user) else flag_key
            )

            for provider in self.flag_check_cache_providers:
                cached_value = provider.get(cache_key)
                if cached_value is not None:
                    return cached_value

            resp = await self.features.check_flag(flag_key, company=company, user=user)
            if resp is None:
                return self._get_flag_default(flag_key)

            for provider in self.flag_check_cache_providers:
                provider.set(cache_key, resp.data.value)

            return resp.data.value
        except Exception as e:
            self.logger.error(e)
            return self._get_flag_default(flag_key)

    async def identify(
        self,
        keys: Dict[str, str],
        company: Optional[EventBodyIdentifyCompany] = None,
        name: Optional[str] = None,
        traits: Optional[Dict[str, Any]] = None,
    ) -> None:
        await self._enqueue_event(
            "identify",
            EventBodyIdentify(
                company=company,
                keys=keys,
                name=name,
                traits=traits,
            ),
        )

    async def track(
        self,
        event: str,
        company: Optional[Dict[str, str]] = None,
        user: Optional[Dict[str, str]] = None,
        traits: Optional[Dict[str, Any]] = None,
    ) -> None:
        await self._enqueue_event(
            "track",
            EventBodyTrack(
                company=company,
                event=event,
                traits=traits,
                user=user,
            ),
        )

    async def _enqueue_event(self, event_type: str, body: EventBody) -> None:
        if self.offline:
            return
        try:
            event_body = CreateEventRequestBody(event_type=event_type, body=body)
            await self.event_buffer.push(event_body)
        except Exception as e:
            self.logger.error(e)

    def _get_flag_default(self, flag_key: str) -> bool:
        return self.flag_defaults.get(flag_key, False)

    def _shutdown_handler(self, signum, frame):
        self.logger.info(f"Received signal {signum}. Initiating shutdown.")
        asyncio.create_task(self.shutdown())

    async def shutdown(self) -> None:
        await self.event_buffer.stop()
        self.logger.info("Shutdown complete.")



# This file was auto-generated by Fern from our API Definition.

import typing

CreateEnvironmentRequestBodyEnvironmentType = typing.Union[
    typing.Literal["development", "staging", "production"], typing.Any
]



# This file was auto-generated by Fern from our API Definition.

from .check_flag_response import CheckFlagResponse
from .check_flags_response import CheckFlagsResponse
from .count_audience_companies_response import CountAudienceCompaniesResponse
from .count_audience_users_response import CountAudienceUsersResponse
from .count_features_params import CountFeaturesParams
from .count_features_response import CountFeaturesResponse
from .count_flag_checks_params import CountFlagChecksParams
from .count_flag_checks_response import CountFlagChecksResponse
from .count_flags_params import CountFlagsParams
from .count_flags_response import CountFlagsResponse
from .create_feature_request_body_feature_type import CreateFeatureRequestBodyFeatureType
from .create_feature_response import CreateFeatureResponse
from .create_flag_response import CreateFlagResponse
from .delete_feature_response import DeleteFeatureResponse
from .delete_flag_response import DeleteFlagResponse
from .get_feature_response import GetFeatureResponse
from .get_flag_check_response import GetFlagCheckResponse
from .get_flag_response import GetFlagResponse
from .get_latest_flag_checks_params import GetLatestFlagChecksParams
from .get_latest_flag_checks_response import GetLatestFlagChecksResponse
from .list_audience_companies_response import ListAudienceCompaniesResponse
from .list_audience_users_response import ListAudienceUsersResponse
from .list_features_params import ListFeaturesParams
from .list_features_response import ListFeaturesResponse
from .list_flag_checks_params import ListFlagChecksParams
from .list_flag_checks_response import ListFlagChecksResponse
from .list_flags_params import ListFlagsParams
from .list_flags_response import ListFlagsResponse
from .update_feature_request_body_feature_type import UpdateFeatureRequestBodyFeatureType
from .update_feature_response import UpdateFeatureResponse
from .update_flag_response import UpdateFlagResponse
from .update_flag_rules_response import UpdateFlagRulesResponse

__all__ = [
    "CheckFlagResponse",
    "CheckFlagsResponse",
    "CountAudienceCompaniesResponse",
    "CountAudienceUsersResponse",
    "CountFeaturesParams",
    "CountFeaturesResponse",
    "CountFlagChecksParams",
    "CountFlagChecksResponse",
    "CountFlagsParams",
    "CountFlagsResponse",
    "CreateFeatureRequestBodyFeatureType",
    "CreateFeatureResponse",
    "CreateFlagResponse",
    "DeleteFeatureResponse",
    "DeleteFlagResponse",
    "GetFeatureResponse",
    "GetFlagCheckResponse",
    "GetFlagResponse",
    "GetLatestFlagChecksParams",
    "GetLatestFlagChecksResponse",
    "ListAudienceCompaniesResponse",
    "ListAudienceUsersResponse",
    "ListFeaturesParams",
    "ListFeaturesResponse",
    "ListFlagChecksParams",
    "ListFlagChecksResponse",
    "ListFlagsParams",
    "ListFlagsResponse",
    "UpdateFeatureRequestBodyFeatureType",
    "UpdateFeatureResponse",
    "UpdateFlagResponse",
    "UpdateFlagRulesResponse",
]



# This file was auto-generated by Fern from our API Definition.

import typing

UpdateEntityTraitDefinitionRequestBodyTraitType = typing.Union[
    typing.Literal["boolean", "currency", "date", "number", "string", "url"], typing.Any
]



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

from ..core.datetime_utils import serialize_datetime
from ..core.pydantic_utilities import deep_union_pydantic_dicts, pydantic_v1
from .feature_detail_response_data import FeatureDetailResponseData
from .plan_response_data import PlanResponseData


class FeatureUsageResponseData(pydantic_v1.BaseModel):
    access: bool = pydantic_v1.Field()
    """
    Whether further usage is permitted.
    """

    allocation: typing.Optional[int] = pydantic_v1.Field(default=None)
    """
    The maximum amount of usage that is permitted; a null value indicates that unlimited usage is permitted.
    """

    entitlement_id: str
    entitlement_type: str
    feature: typing.Optional[FeatureDetailResponseData] = None
    period: typing.Optional[str] = pydantic_v1.Field(default=None)
    """
    The period over which usage is measured.
    """

    plan: typing.Optional[PlanResponseData] = None
    usage: typing.Optional[int] = pydantic_v1.Field(default=None)
    """
    The amount of usage that has been consumed; a null value indicates that usage is not being measured.
    """

    def json(self, **kwargs: typing.Any) -> str:
        kwargs_with_defaults: typing.Any = {"by_alias": True, "exclude_unset": True, **kwargs}
        return super().json(**kwargs_with_defaults)

    def dict(self, **kwargs: typing.Any) -> typing.Dict[str, typing.Any]:
        kwargs_with_defaults_exclude_unset: typing.Any = {"by_alias": True, "exclude_unset": True, **kwargs}
        kwargs_with_defaults_exclude_none: typing.Any = {"by_alias": True, "exclude_none": True, **kwargs}

        return deep_union_pydantic_dicts(
            super().dict(**kwargs_with_defaults_exclude_unset), super().dict(**kwargs_with_defaults_exclude_none)
        )

    class Config:
        frozen = True
        smart_union = True
        extra = pydantic_v1.Extra.allow
        json_encoders = {dt.datetime: serialize_datetime}



import logging
from typing import Optional


def get_default_logger(name: str = "schematic") -> logging.Logger:
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    if not logger.handlers:
        # Create console handler
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)

        # Create formatter and add it to the handlers
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        ch.setFormatter(formatter)

        # Add the handler to the logger
        logger.addHandler(ch)

    return logger



# This file was auto-generated by Fern from our API Definition.

import typing

UpdatePlanEntitlementRequestBodyValueType = typing.Union[
    typing.Literal["Boolean", "Numeric", "Trait", "Unlimited"], typing.Any
]



# This file was auto-generated by Fern from our API Definition.

from .types import (
    ApiError,
    ApiKeyCreateResponseData,
    ApiKeyRequestListResponseData,
    ApiKeyRequestResponseData,
    ApiKeyResponseData,
    AudienceRequestBody,
    BillingProductResponseData,
    BillingSubscriptionResponseData,
    CheckFlagOutputWithFlagKey,
    CheckFlagRequestBody,
    CheckFlagResponseData,
    CheckFlagsResponseData,
    CompanyCrmDealsResponseData,
    CompanyDetailResponseData,
    CompanyMembershipDetailResponseData,
    CompanyMembershipResponseData,
    CompanyOverrideResponseData,
    CompanyPlanResponseData,
    CompanyResponseData,
    CompanySubscriptionResponseData,
    CountResponse,
    CreateEventRequestBody,
    CreateEventRequestBodyEventType,
    CreateFlagRequestBody,
    CreateOrUpdateConditionGroupRequestBody,
    CreateOrUpdateConditionRequestBody,
    CreateOrUpdateConditionRequestBodyConditionType,
    CreateOrUpdateConditionRequestBodyMetricPeriod,
    CreateOrUpdateConditionRequestBodyOperator,
    CreateOrUpdateFlagRequestBody,
    CreateOrUpdateRuleRequestBody,
    CreateOrUpdateRuleRequestBodyRuleType,
    CreateReqCommon,
    CreateReqCommonMetricPeriod,
    CreateReqCommonValueType,
    CrmDealLineItem,
    CrmDealResponseData,
    CrmLineItemResponseData,
    CrmProductResponseData,
    Decimal,
    DeleteResponse,
    EntityKeyDefinitionResponseData,
    EntityKeyDetailResponseData,
    EntityKeyResponseData,
    EntityTraitDefinitionResponseData,
    EntityTraitDetailResponseData,
    EntityTraitResponseData,
    EntityTraitValue,
    EnvironmentDetailResponseData,
    EnvironmentResponseData,
    EventBody,
    EventBodyIdentify,
    EventBodyIdentifyCompany,
    EventBodyTrack,
    EventDetailResponseData,
    EventResponseData,
    EventSummaryResponseData,
    FeatureCompanyResponseData,
    FeatureCompanyUserResponseData,
    FeatureDetailResponseData,
    FeatureResponseData,
    FeatureUsageDetailResponseData,
    FeatureUsageResponseData,
    FlagCheckLogDetailResponseData,
    FlagCheckLogResponseData,
    FlagDetailResponseData,
    FlagResponseData,
    KeysRequestBody,
    MetricCountsHourlyResponseData,
    PaginationFilter,
    PlanAudienceDetailResponseData,
    PlanAudienceResponseData,
    PlanDetailResponseData,
    PlanEntitlementResponseData,
    PlanResponseData,
    PreviewObject,
    RawEventBatchResponseData,
    RawEventResponseData,
    RuleConditionDetailResponseData,
    RuleConditionGroupDetailResponseData,
    RuleConditionGroupResponseData,
    RuleConditionResourceResponseData,
    RuleConditionResponseData,
    RuleDetailResponseData,
    RuleResponseData,
    RulesDetailResponseData,
    SegmentStatusResp,
    UpdateReqCommon,
    UpdateReqCommonMetricPeriod,
    UpdateReqCommonValueType,
    UpdateRuleRequestBody,
    UpsertCompanyRequestBody,
    UpsertTraitRequestBody,
    UpsertUserRequestBody,
    UpsertUserSubRequestBody,
    UserDetailResponseData,
    UserResponseData,
    WebhookEventDetailResponseData,
    WebhookEventResponseData,
    WebhookResponseData,
)
from .errors import BadRequestError, ForbiddenError, InternalServerError, NotFoundError, UnauthorizedError
from . import accounts, billing, companies, crm, entitlements, events, features, plans, webhooks
from .accounts import (
    CountApiKeysParams,
    CountApiKeysResponse,
    CountApiRequestsParams,
    CountApiRequestsResponse,
    CreateApiKeyResponse,
    CreateEnvironmentRequestBodyEnvironmentType,
    CreateEnvironmentResponse,
    DeleteApiKeyResponse,
    DeleteEnvironmentResponse,
    GetApiKeyResponse,
    GetApiRequestResponse,
    GetEnvironmentResponse,
    ListApiKeysParams,
    ListApiKeysResponse,
    ListApiRequestsParams,
    ListApiRequestsResponse,
    ListEnvironmentsParams,
    ListEnvironmentsResponse,
    UpdateApiKeyResponse,
    UpdateEnvironmentRequestBodyEnvironmentType,
    UpdateEnvironmentResponse,
)
from .billing import (
    ListProductsParams,
    ListProductsResponse,
    UpsertBillingProductResponse,
    UpsertBillingSubscriptionResponse,
)
from .client import AsyncSchematic, AsyncSchematicConfig, LocalCache, Schematic, SchematicConfig
from .companies import (
    CountCompaniesParams,
    CountCompaniesResponse,
    CountEntityKeyDefinitionsParams,
    CountEntityKeyDefinitionsResponse,
    CountEntityTraitDefinitionsParams,
    CountEntityTraitDefinitionsResponse,
    CountUsersParams,
    CountUsersResponse,
    CreateCompanyResponse,
    CreateEntityTraitDefinitionRequestBodyEntityType,
    CreateEntityTraitDefinitionRequestBodyTraitType,
    CreateUserResponse,
    DeleteCompanyByKeysResponse,
    DeleteCompanyMembershipResponse,
    DeleteCompanyResponse,
    DeleteUserByKeysResponse,
    DeleteUserResponse,
    GetActiveCompanySubscriptionParams,
    GetActiveCompanySubscriptionResponse,
    GetActiveDealsParams,
    GetActiveDealsResponse,
    GetCompanyResponse,
    GetEntityTraitDefinitionResponse,
    GetEntityTraitValuesParams,
    GetEntityTraitValuesResponse,
    GetOrCreateCompanyMembershipResponse,
    GetOrCreateEntityTraitDefinitionResponse,
    GetUserResponse,
    ListCompaniesParams,
    ListCompaniesResponse,
    ListCompanyMembershipsParams,
    ListCompanyMembershipsResponse,
    ListCompanyPlansParams,
    ListCompanyPlansResponse,
    ListEntityKeyDefinitionsParams,
    ListEntityKeyDefinitionsResponse,
    ListEntityTraitDefinitionsParams,
    ListEntityTraitDefinitionsResponse,
    ListUsersParams,
    ListUsersResponse,
    LookupCompanyParams,
    LookupCompanyResponse,
    LookupUserParams,
    LookupUserResponse,
    UpdateEntityTraitDefinitionRequestBodyTraitType,
    UpdateEntityTraitDefinitionResponse,
    UpsertCompanyResponse,
    UpsertCompanyTraitResponse,
    UpsertUserResponse,
    UpsertUserTraitResponse,
)
from .crm import (
    ListCrmProductsParams,
    ListCrmProductsResponse,
    UpsertCrmDealResponse,
    UpsertCrmProductResponse,
    UpsertDealLineItemAssociationResponse,
    UpsertLineItemResponse,
)
from .entitlements import (
    CountCompanyOverridesParams,
    CountCompanyOverridesResponse,
    CountFeatureCompaniesParams,
    CountFeatureCompaniesResponse,
    CountFeatureUsageParams,
    CountFeatureUsageResponse,
    CountFeatureUsersParams,
    CountFeatureUsersResponse,
    CountPlanEntitlementsParams,
    CountPlanEntitlementsResponse,
    CreateCompanyOverrideRequestBodyMetricPeriod,
    CreateCompanyOverrideRequestBodyValueType,
    CreateCompanyOverrideResponse,
    CreatePlanEntitlementRequestBodyMetricPeriod,
    CreatePlanEntitlementRequestBodyValueType,
    CreatePlanEntitlementResponse,
    DeleteCompanyOverrideResponse,
    DeletePlanEntitlementResponse,
    GetCompanyOverrideResponse,
    GetFeatureUsageByCompanyParams,
    GetFeatureUsageByCompanyResponse,
    GetPlanEntitlementResponse,
    ListCompanyOverridesParams,
    ListCompanyOverridesResponse,
    ListFeatureCompaniesParams,
    ListFeatureCompaniesResponse,
    ListFeatureUsageParams,
    ListFeatureUsageResponse,
    ListFeatureUsersParams,
    ListFeatureUsersResponse,
    ListPlanEntitlementsParams,
    ListPlanEntitlementsResponse,
    UpdateCompanyOverrideRequestBodyMetricPeriod,
    UpdateCompanyOverrideRequestBodyValueType,
    UpdateCompanyOverrideResponse,
    UpdatePlanEntitlementRequestBodyMetricPeriod,
    UpdatePlanEntitlementRequestBodyValueType,
    UpdatePlanEntitlementResponse,
)
from .environment import SchematicEnvironment
from .events import (
    CreateEventBatchResponse,
    CreateEventResponse,
    GetEventResponse,
    GetEventSummariesParams,
    GetEventSummariesResponse,
    GetEventSummaryBySubtypeResponse,
    GetSegmentIntegrationStatusResponse,
    ListEventsParams,
    ListEventsResponse,
    ListMetricCountsParams,
    ListMetricCountsResponse,
)
from .features import (
    CheckFlagResponse,
    CheckFlagsResponse,
    CountAudienceCompaniesResponse,
    CountAudienceUsersResponse,
    CountFeaturesParams,
    CountFeaturesResponse,
    CountFlagChecksParams,
    CountFlagChecksResponse,
    CountFlagsParams,
    CountFlagsResponse,
    CreateFeatureRequestBodyFeatureType,
    CreateFeatureResponse,
    CreateFlagResponse,
    DeleteFeatureResponse,
    DeleteFlagResponse,
    GetFeatureResponse,
    GetFlagCheckResponse,
    GetFlagResponse,
    GetLatestFlagChecksParams,
    GetLatestFlagChecksResponse,
    ListAudienceCompaniesResponse,
    ListAudienceUsersResponse,
    ListFeaturesParams,
    ListFeaturesResponse,
    ListFlagChecksParams,
    ListFlagChecksResponse,
    ListFlagsParams,
    ListFlagsResponse,
    UpdateFeatureRequestBodyFeatureType,
    UpdateFeatureResponse,
    UpdateFlagResponse,
    UpdateFlagRulesResponse,
)
from .plans import (
    CountPlansParams,
    CountPlansResponse,
    CreatePlanRequestBodyPlanType,
    CreatePlanResponse,
    DeleteAudienceResponse,
    DeletePlanResponse,
    GetAudienceResponse,
    GetPlanResponse,
    ListPlansParams,
    ListPlansResponse,
    UpdateAudienceResponse,
    UpdatePlanRequestBodyPlanType,
    UpdatePlanResponse,
)
from .version import __version__
from .webhooks import (
    CountWebhookEventsParams,
    CountWebhookEventsResponse,
    CountWebhooksParams,
    CountWebhooksResponse,
    CreateWebhookResponse,
    DeleteWebhookResponse,
    GetWebhookEventResponse,
    GetWebhookResponse,
    ListWebhookEventsParams,
    ListWebhookEventsResponse,
    ListWebhooksParams,
    ListWebhooksResponse,
    UpdateWebhookResponse,
)

__all__ = [
    "ApiError",
    "ApiKeyCreateResponseData",
    "ApiKeyRequestListResponseData",
    "ApiKeyRequestResponseData",
    "ApiKeyResponseData",
    "AsyncSchematic",
    "AsyncSchematicConfig",
    "AudienceRequestBody",
    "BadRequestError",
    "BillingProductResponseData",
    "BillingSubscriptionResponseData",
    "CheckFlagOutputWithFlagKey",
    "CheckFlagRequestBody",
    "CheckFlagResponse",
    "CheckFlagResponseData",
    "CheckFlagsResponse",
    "CheckFlagsResponseData",
    "CompanyCrmDealsResponseData",
    "CompanyDetailResponseData",
    "CompanyMembershipDetailResponseData",
    "CompanyMembershipResponseData",
    "CompanyOverrideResponseData",
    "CompanyPlanResponseData",
    "CompanyResponseData",
    "CompanySubscriptionResponseData",
    "CountApiKeysParams",
    "CountApiKeysResponse",
    "CountApiRequestsParams",
    "CountApiRequestsResponse",
    "CountAudienceCompaniesResponse",
    "CountAudienceUsersResponse",
    "CountCompaniesParams",
    "CountCompaniesResponse",
    "CountCompanyOverridesParams",
    "CountCompanyOverridesResponse",
    "CountEntityKeyDefinitionsParams",
    "CountEntityKeyDefinitionsResponse",
    "CountEntityTraitDefinitionsParams",
    "CountEntityTraitDefinitionsResponse",
    "CountFeatureCompaniesParams",
    "CountFeatureCompaniesResponse",
    "CountFeatureUsageParams",
    "CountFeatureUsageResponse",
    "CountFeatureUsersParams",
    "CountFeatureUsersResponse",
    "CountFeaturesParams",
    "CountFeaturesResponse",
    "CountFlagChecksParams",
    "CountFlagChecksResponse",
    "CountFlagsParams",
    "CountFlagsResponse",
    "CountPlanEntitlementsParams",
    "CountPlanEntitlementsResponse",
    "CountPlansParams",
    "CountPlansResponse",
    "CountResponse",
    "CountUsersParams",
    "CountUsersResponse",
    "CountWebhookEventsParams",
    "CountWebhookEventsResponse",
    "CountWebhooksParams",
    "CountWebhooksResponse",
    "CreateApiKeyResponse",
    "CreateCompanyOverrideRequestBodyMetricPeriod",
    "CreateCompanyOverrideRequestBodyValueType",
    "CreateCompanyOverrideResponse",
    "CreateCompanyResponse",
    "CreateEntityTraitDefinitionRequestBodyEntityType",
    "CreateEntityTraitDefinitionRequestBodyTraitType",
    "CreateEnvironmentRequestBodyEnvironmentType",
    "CreateEnvironmentResponse",
    "CreateEventBatchResponse",
    "CreateEventRequestBody",
    "CreateEventRequestBodyEventType",
    "CreateEventResponse",
    "CreateFeatureRequestBodyFeatureType",
    "CreateFeatureResponse",
    "CreateFlagRequestBody",
    "CreateFlagResponse",
    "CreateOrUpdateConditionGroupRequestBody",
    "CreateOrUpdateConditionRequestBody",
    "CreateOrUpdateConditionRequestBodyConditionType",
    "CreateOrUpdateConditionRequestBodyMetricPeriod",
    "CreateOrUpdateConditionRequestBodyOperator",
    "CreateOrUpdateFlagRequestBody",
    "CreateOrUpdateRuleRequestBody",
    "CreateOrUpdateRuleRequestBodyRuleType",
    "CreatePlanEntitlementRequestBodyMetricPeriod",
    "CreatePlanEntitlementRequestBodyValueType",
    "CreatePlanEntitlementResponse",
    "CreatePlanRequestBodyPlanType",
    "CreatePlanResponse",
    "CreateReqCommon",
    "CreateReqCommonMetricPeriod",
    "CreateReqCommonValueType",
    "CreateUserResponse",
    "CreateWebhookResponse",
    "CrmDealLineItem",
    "CrmDealResponseData",
    "CrmLineItemResponseData",
    "CrmProductResponseData",
    "Decimal",
    "DeleteApiKeyResponse",
    "DeleteAudienceResponse",
    "DeleteCompanyByKeysResponse",
    "DeleteCompanyMembershipResponse",
    "DeleteCompanyOverrideResponse",
    "DeleteCompanyResponse",
    "DeleteEnvironmentResponse",
    "DeleteFeatureResponse",
    "DeleteFlagResponse",
    "DeletePlanEntitlementResponse",
    "DeletePlanResponse",
    "DeleteResponse",
    "DeleteUserByKeysResponse",
    "DeleteUserResponse",
    "DeleteWebhookResponse",
    "EntityKeyDefinitionResponseData",
    "EntityKeyDetailResponseData",
    "EntityKeyResponseData",
    "EntityTraitDefinitionResponseData",
    "EntityTraitDetailResponseData",
    "EntityTraitResponseData",
    "EntityTraitValue",
    "EnvironmentDetailResponseData",
    "EnvironmentResponseData",
    "EventBody",
    "EventBodyIdentify",
    "EventBodyIdentifyCompany",
    "EventBodyTrack",
    "EventDetailResponseData",
    "EventResponseData",
    "EventSummaryResponseData",
    "FeatureCompanyResponseData",
    "FeatureCompanyUserResponseData",
    "FeatureDetailResponseData",
    "FeatureResponseData",
    "FeatureUsageDetailResponseData",
    "FeatureUsageResponseData",
    "FlagCheckLogDetailResponseData",
    "FlagCheckLogResponseData",
    "FlagDetailResponseData",
    "FlagResponseData",
    "ForbiddenError",
    "GetActiveCompanySubscriptionParams",
    "GetActiveCompanySubscriptionResponse",
    "GetActiveDealsParams",
    "GetActiveDealsResponse",
    "GetApiKeyResponse",
    "GetApiRequestResponse",
    "GetAudienceResponse",
    "GetCompanyOverrideResponse",
    "GetCompanyResponse",
    "GetEntityTraitDefinitionResponse",
    "GetEntityTraitValuesParams",
    "GetEntityTraitValuesResponse",
    "GetEnvironmentResponse",
    "GetEventResponse",
    "GetEventSummariesParams",
    "GetEventSummariesResponse",
    "GetEventSummaryBySubtypeResponse",
    "GetFeatureResponse",
    "GetFeatureUsageByCompanyParams",
    "GetFeatureUsageByCompanyResponse",
    "GetFlagCheckResponse",
    "GetFlagResponse",
    "GetLatestFlagChecksParams",
    "GetLatestFlagChecksResponse",
    "GetOrCreateCompanyMembershipResponse",
    "GetOrCreateEntityTraitDefinitionResponse",
    "GetPlanEntitlementResponse",
    "GetPlanResponse",
    "GetSegmentIntegrationStatusResponse",
    "GetUserResponse",
    "GetWebhookEventResponse",
    "GetWebhookResponse",
    "InternalServerError",
    "KeysRequestBody",
    "ListApiKeysParams",
    "ListApiKeysResponse",
    "ListApiRequestsParams",
    "ListApiRequestsResponse",
    "ListAudienceCompaniesResponse",
    "ListAudienceUsersResponse",
    "ListCompaniesParams",
    "ListCompaniesResponse",
    "ListCompanyMembershipsParams",
    "ListCompanyMembershipsResponse",
    "ListCompanyOverridesParams",
    "ListCompanyOverridesResponse",
    "ListCompanyPlansParams",
    "ListCompanyPlansResponse",
    "ListCrmProductsParams",
    "ListCrmProductsResponse",
    "ListEntityKeyDefinitionsParams",
    "ListEntityKeyDefinitionsResponse",
    "ListEntityTraitDefinitionsParams",
    "ListEntityTraitDefinitionsResponse",
    "ListEnvironmentsParams",
    "ListEnvironmentsResponse",
    "ListEventsParams",
    "ListEventsResponse",
    "ListFeatureCompaniesParams",
    "ListFeatureCompaniesResponse",
    "ListFeatureUsageParams",
    "ListFeatureUsageResponse",
    "ListFeatureUsersParams",
    "ListFeatureUsersResponse",
    "ListFeaturesParams",
    "ListFeaturesResponse",
    "ListFlagChecksParams",
    "ListFlagChecksResponse",
    "ListFlagsParams",
    "ListFlagsResponse",
    "ListMetricCountsParams",
    "ListMetricCountsResponse",
    "ListPlanEntitlementsParams",
    "ListPlanEntitlementsResponse",
    "ListPlansParams",
    "ListPlansResponse",
    "ListProductsParams",
    "ListProductsResponse",
    "ListUsersParams",
    "ListUsersResponse",
    "ListWebhookEventsParams",
    "ListWebhookEventsResponse",
    "ListWebhooksParams",
    "ListWebhooksResponse",
    "LocalCache",
    "LookupCompanyParams",
    "LookupCompanyResponse",
    "LookupUserParams",
    "LookupUserResponse",
    "MetricCountsHourlyResponseData",
    "NotFoundError",
    "PaginationFilter",
    "PlanAudienceDetailResponseData",
    "PlanAudienceResponseData",
    "PlanDetailResponseData",
    "PlanEntitlementResponseData",
    "PlanResponseData",
    "PreviewObject",
    "RawEventBatchResponseData",
    "RawEventResponseData",
    "RuleConditionDetailResponseData",
    "RuleConditionGroupDetailResponseData",
    "RuleConditionGroupResponseData",
    "RuleConditionResourceResponseData",
    "RuleConditionResponseData",
    "RuleDetailResponseData",
    "RuleResponseData",
    "RulesDetailResponseData",
    "Schematic",
    "SchematicConfig",
    "SchematicEnvironment",
    "SegmentStatusResp",
    "UnauthorizedError",
    "UpdateApiKeyResponse",
    "UpdateAudienceResponse",
    "UpdateCompanyOverrideRequestBodyMetricPeriod",
    "UpdateCompanyOverrideRequestBodyValueType",
    "UpdateCompanyOverrideResponse",
    "UpdateEntityTraitDefinitionRequestBodyTraitType",
    "UpdateEntityTraitDefinitionResponse",
    "UpdateEnvironmentRequestBodyEnvironmentType",
    "UpdateEnvironmentResponse",
    "UpdateFeatureRequestBodyFeatureType",
    "UpdateFeatureResponse",
    "UpdateFlagResponse",
    "UpdateFlagRulesResponse",
    "UpdatePlanEntitlementRequestBodyMetricPeriod",
    "UpdatePlanEntitlementRequestBodyValueType",
    "UpdatePlanEntitlementResponse",
    "UpdatePlanRequestBodyPlanType",
    "UpdatePlanResponse",
    "UpdateReqCommon",
    "UpdateReqCommonMetricPeriod",
    "UpdateReqCommonValueType",
    "UpdateRuleRequestBody",
    "UpdateWebhookResponse",
    "UpsertBillingProductResponse",
    "UpsertBillingSubscriptionResponse",
    "UpsertCompanyRequestBody",
    "UpsertCompanyResponse",
    "UpsertCompanyTraitResponse",
    "UpsertCrmDealResponse",
    "UpsertCrmProductResponse",
    "UpsertDealLineItemAssociationResponse",
    "UpsertLineItemResponse",
    "UpsertTraitRequestBody",
    "UpsertUserRequestBody",
    "UpsertUserResponse",
    "UpsertUserSubRequestBody",
    "UpsertUserTraitResponse",
    "UserDetailResponseData",
    "UserResponseData",
    "WebhookEventDetailResponseData",
    "WebhookEventResponseData",
    "WebhookResponseData",
    "__version__",
    "accounts",
    "billing",
    "companies",
    "crm",
    "entitlements",
    "events",
    "features",
    "plans",
    "webhooks",
]



# This file was auto-generated by Fern from our API Definition.

from .list_crm_products_params import ListCrmProductsParams
from .list_crm_products_response import ListCrmProductsResponse
from .upsert_crm_deal_response import UpsertCrmDealResponse
from .upsert_crm_product_response import UpsertCrmProductResponse
from .upsert_deal_line_item_association_response import UpsertDealLineItemAssociationResponse
from .upsert_line_item_response import UpsertLineItemResponse

__all__ = [
    "ListCrmProductsParams",
    "ListCrmProductsResponse",
    "UpsertCrmDealResponse",
    "UpsertCrmProductResponse",
    "UpsertDealLineItemAssociationResponse",
    "UpsertLineItemResponse",
]



# This file was auto-generated by Fern from our API Definition.

import typing

CreateOrUpdateConditionRequestBodyConditionType = typing.Union[
    typing.Literal["company", "metric", "trait", "user", "plan", "billing_product", "crm_product"], typing.Any
]



# This file was auto-generated by Fern from our API Definition.

import typing

UpdateCompanyOverrideRequestBodyMetricPeriod = typing.Union[
    typing.Literal["current_month", "current_week", "current_day"], typing.Any
]



# This file was auto-generated by Fern from our API Definition.

import typing

CreateOrUpdateRuleRequestBodyRuleType = typing.Union[
    typing.Literal["global_override", "company_override", "plan_entitlement", "standard", "default", "plan_audience"],
    typing.Any,
]



# This file was auto-generated by Fern from our API Definition.

from .count_plans_params import CountPlansParams
from .count_plans_response import CountPlansResponse
from .create_plan_request_body_plan_type import CreatePlanRequestBodyPlanType
from .create_plan_response import CreatePlanResponse
from .delete_audience_response import DeleteAudienceResponse
from .delete_plan_response import DeletePlanResponse
from .get_audience_response import GetAudienceResponse
from .get_plan_response import GetPlanResponse
from .list_plans_params import ListPlansParams
from .list_plans_response import ListPlansResponse
from .update_audience_response import UpdateAudienceResponse
from .update_plan_request_body_plan_type import UpdatePlanRequestBodyPlanType
from .update_plan_response import UpdatePlanResponse

__all__ = [
    "CountPlansParams",
    "CountPlansResponse",
    "CreatePlanRequestBodyPlanType",
    "CreatePlanResponse",
    "DeleteAudienceResponse",
    "DeletePlanResponse",
    "GetAudienceResponse",
    "GetPlanResponse",
    "ListPlansParams",
    "ListPlansResponse",
    "UpdateAudienceResponse",
    "UpdatePlanRequestBodyPlanType",
    "UpdatePlanResponse",
]



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

from ..core.datetime_utils import serialize_datetime
from ..core.pydantic_utilities import deep_union_pydantic_dicts, pydantic_v1
from .entity_trait_definition_response_data import EntityTraitDefinitionResponseData
from .rule_condition_resource_response_data import RuleConditionResourceResponseData


class RuleConditionDetailResponseData(pydantic_v1.BaseModel):
    comparison_trait: typing.Optional[EntityTraitDefinitionResponseData] = None
    comparison_trait_id: typing.Optional[str] = None
    condition_group_id: typing.Optional[str] = None
    condition_type: str
    created_at: dt.datetime
    environment_id: str
    event_subtype: typing.Optional[str] = None
    flag_id: typing.Optional[str] = None
    id: str
    metric_period: typing.Optional[str] = None
    metric_value: int
    operator: str
    plan_id: typing.Optional[str] = None
    resource_ids: typing.List[str]
    resources: typing.List[RuleConditionResourceResponseData]
    rule_id: str
    trait: typing.Optional[EntityTraitDefinitionResponseData] = None
    trait_entity_type: typing.Optional[str] = None
    trait_id: typing.Optional[str] = None
    trait_value: str
    updated_at: dt.datetime

    def json(self, **kwargs: typing.Any) -> str:
        kwargs_with_defaults: typing.Any = {"by_alias": True, "exclude_unset": True, **kwargs}
        return super().json(**kwargs_with_defaults)

    def dict(self, **kwargs: typing.Any) -> typing.Dict[str, typing.Any]:
        kwargs_with_defaults_exclude_unset: typing.Any = {"by_alias": True, "exclude_unset": True, **kwargs}
        kwargs_with_defaults_exclude_none: typing.Any = {"by_alias": True, "exclude_none": True, **kwargs}

        return deep_union_pydantic_dicts(
            super().dict(**kwargs_with_defaults_exclude_unset), super().dict(**kwargs_with_defaults_exclude_none)
        )

    class Config:
        frozen = True
        smart_union = True
        extra = pydantic_v1.Extra.allow
        json_encoders = {dt.datetime: serialize_datetime}



# This file was auto-generated by Fern from our API Definition.

from .count_companies_params import CountCompaniesParams
from .count_companies_response import CountCompaniesResponse
from .count_entity_key_definitions_params import CountEntityKeyDefinitionsParams
from .count_entity_key_definitions_response import CountEntityKeyDefinitionsResponse
from .count_entity_trait_definitions_params import CountEntityTraitDefinitionsParams
from .count_entity_trait_definitions_response import CountEntityTraitDefinitionsResponse
from .count_users_params import CountUsersParams
from .count_users_response import CountUsersResponse
from .create_company_response import CreateCompanyResponse
from .create_entity_trait_definition_request_body_entity_type import CreateEntityTraitDefinitionRequestBodyEntityType
from .create_entity_trait_definition_request_body_trait_type import CreateEntityTraitDefinitionRequestBodyTraitType
from .create_user_response import CreateUserResponse
from .delete_company_by_keys_response import DeleteCompanyByKeysResponse
from .delete_company_membership_response import DeleteCompanyMembershipResponse
from .delete_company_response import DeleteCompanyResponse
from .delete_user_by_keys_response import DeleteUserByKeysResponse
from .delete_user_response import DeleteUserResponse
from .get_active_company_subscription_params import GetActiveCompanySubscriptionParams
from .get_active_company_subscription_response import GetActiveCompanySubscriptionResponse
from .get_active_deals_params import GetActiveDealsParams
from .get_active_deals_response import GetActiveDealsResponse
from .get_company_response import GetCompanyResponse
from .get_entity_trait_definition_response import GetEntityTraitDefinitionResponse
from .get_entity_trait_values_params import GetEntityTraitValuesParams
from .get_entity_trait_values_response import GetEntityTraitValuesResponse
from .get_or_create_company_membership_response import GetOrCreateCompanyMembershipResponse
from .get_or_create_entity_trait_definition_response import GetOrCreateEntityTraitDefinitionResponse
from .get_user_response import GetUserResponse
from .list_companies_params import ListCompaniesParams
from .list_companies_response import ListCompaniesResponse
from .list_company_memberships_params import ListCompanyMembershipsParams
from .list_company_memberships_response import ListCompanyMembershipsResponse
from .list_company_plans_params import ListCompanyPlansParams
from .list_company_plans_response import ListCompanyPlansResponse
from .list_entity_key_definitions_params import ListEntityKeyDefinitionsParams
from .list_entity_key_definitions_response import ListEntityKeyDefinitionsResponse
from .list_entity_trait_definitions_params import ListEntityTraitDefinitionsParams
from .list_entity_trait_definitions_response import ListEntityTraitDefinitionsResponse
from .list_users_params import ListUsersParams
from .list_users_response import ListUsersResponse
from .lookup_company_params import LookupCompanyParams
from .lookup_company_response import LookupCompanyResponse
from .lookup_user_params import LookupUserParams
from .lookup_user_response import LookupUserResponse
from .update_entity_trait_definition_request_body_trait_type import UpdateEntityTraitDefinitionRequestBodyTraitType
from .update_entity_trait_definition_response import UpdateEntityTraitDefinitionResponse
from .upsert_company_response import UpsertCompanyResponse
from .upsert_company_trait_response import UpsertCompanyTraitResponse
from .upsert_user_response import UpsertUserResponse
from .upsert_user_trait_response import UpsertUserTraitResponse

__all__ = [
    "CountCompaniesParams",
    "CountCompaniesResponse",
    "CountEntityKeyDefinitionsParams",
    "CountEntityKeyDefinitionsResponse",
    "CountEntityTraitDefinitionsParams",
    "CountEntityTraitDefinitionsResponse",
    "CountUsersParams",
    "CountUsersResponse",
    "CreateCompanyResponse",
    "CreateEntityTraitDefinitionRequestBodyEntityType",
    "CreateEntityTraitDefinitionRequestBodyTraitType",
    "CreateUserResponse",
    "DeleteCompanyByKeysResponse",
    "DeleteCompanyMembershipResponse",
    "DeleteCompanyResponse",
    "DeleteUserByKeysResponse",
    "DeleteUserResponse",
    "GetActiveCompanySubscriptionParams",
    "GetActiveCompanySubscriptionResponse",
    "GetActiveDealsParams",
    "GetActiveDealsResponse",
    "GetCompanyResponse",
    "GetEntityTraitDefinitionResponse",
    "GetEntityTraitValuesParams",
    "GetEntityTraitValuesResponse",
    "GetOrCreateCompanyMembershipResponse",
    "GetOrCreateEntityTraitDefinitionResponse",
    "GetUserResponse",
    "ListCompaniesParams",
    "ListCompaniesResponse",
    "ListCompanyMembershipsParams",
    "ListCompanyMembershipsResponse",
    "ListCompanyPlansParams",
    "ListCompanyPlansResponse",
    "ListEntityKeyDefinitionsParams",
    "ListEntityKeyDefinitionsResponse",
    "ListEntityTraitDefinitionsParams",
    "ListEntityTraitDefinitionsResponse",
    "ListUsersParams",
    "ListUsersResponse",
    "LookupCompanyParams",
    "LookupCompanyResponse",
    "LookupUserParams",
    "LookupUserResponse",
    "UpdateEntityTraitDefinitionRequestBodyTraitType",
    "UpdateEntityTraitDefinitionResponse",
    "UpsertCompanyResponse",
    "UpsertCompanyTraitResponse",
    "UpsertUserResponse",
    "UpsertUserTraitResponse",
]



# This file was auto-generated by Fern from our API Definition.

from .list_products_params import ListProductsParams
from .list_products_response import ListProductsResponse
from .upsert_billing_product_response import UpsertBillingProductResponse
from .upsert_billing_subscription_response import UpsertBillingSubscriptionResponse

__all__ = [
    "ListProductsParams",
    "ListProductsResponse",
    "UpsertBillingProductResponse",
    "UpsertBillingSubscriptionResponse",
]



# This file was auto-generated by Fern from our API Definition.

from .types import (
    CountCompanyOverridesParams,
    CountCompanyOverridesResponse,
    CountFeatureCompaniesParams,
    CountFeatureCompaniesResponse,
    CountFeatureUsageParams,
    CountFeatureUsageResponse,
    CountFeatureUsersParams,
    CountFeatureUsersResponse,
    CountPlanEntitlementsParams,
    CountPlanEntitlementsResponse,
    CreateCompanyOverrideRequestBodyMetricPeriod,
    CreateCompanyOverrideRequestBodyValueType,
    CreateCompanyOverrideResponse,
    CreatePlanEntitlementRequestBodyMetricPeriod,
    CreatePlanEntitlementRequestBodyValueType,
    CreatePlanEntitlementResponse,
    DeleteCompanyOverrideResponse,
    DeletePlanEntitlementResponse,
    GetCompanyOverrideResponse,
    GetFeatureUsageByCompanyParams,
    GetFeatureUsageByCompanyResponse,
    GetPlanEntitlementResponse,
    ListCompanyOverridesParams,
    ListCompanyOverridesResponse,
    ListFeatureCompaniesParams,
    ListFeatureCompaniesResponse,
    ListFeatureUsageParams,
    ListFeatureUsageResponse,
    ListFeatureUsersParams,
    ListFeatureUsersResponse,
    ListPlanEntitlementsParams,
    ListPlanEntitlementsResponse,
    UpdateCompanyOverrideRequestBodyMetricPeriod,
    UpdateCompanyOverrideRequestBodyValueType,
    UpdateCompanyOverrideResponse,
    UpdatePlanEntitlementRequestBodyMetricPeriod,
    UpdatePlanEntitlementRequestBodyValueType,
    UpdatePlanEntitlementResponse,
)

__all__ = [
    "CountCompanyOverridesParams",
    "CountCompanyOverridesResponse",
    "CountFeatureCompaniesParams",
    "CountFeatureCompaniesResponse",
    "CountFeatureUsageParams",
    "CountFeatureUsageResponse",
    "CountFeatureUsersParams",
    "CountFeatureUsersResponse",
    "CountPlanEntitlementsParams",
    "CountPlanEntitlementsResponse",
    "CreateCompanyOverrideRequestBodyMetricPeriod",
    "CreateCompanyOverrideRequestBodyValueType",
    "CreateCompanyOverrideResponse",
    "CreatePlanEntitlementRequestBodyMetricPeriod",
    "CreatePlanEntitlementRequestBodyValueType",
    "CreatePlanEntitlementResponse",
    "DeleteCompanyOverrideResponse",
    "DeletePlanEntitlementResponse",
    "GetCompanyOverrideResponse",
    "GetFeatureUsageByCompanyParams",
    "GetFeatureUsageByCompanyResponse",
    "GetPlanEntitlementResponse",
    "ListCompanyOverridesParams",
    "ListCompanyOverridesResponse",
    "ListFeatureCompaniesParams",
    "ListFeatureCompaniesResponse",
    "ListFeatureUsageParams",
    "ListFeatureUsageResponse",
    "ListFeatureUsersParams",
    "ListFeatureUsersResponse",
    "ListPlanEntitlementsParams",
    "ListPlanEntitlementsResponse",
    "UpdateCompanyOverrideRequestBodyMetricPeriod",
    "UpdateCompanyOverrideRequestBodyValueType",
    "UpdateCompanyOverrideResponse",
    "UpdatePlanEntitlementRequestBodyMetricPeriod",
    "UpdatePlanEntitlementRequestBodyValueType",
    "UpdatePlanEntitlementResponse",
]



import httpx


class OfflineHTTPClient(httpx.Client):
    def send(self, request, **kwargs):
        response = httpx.Response(
            status_code=200,
            json={"data": {}},
            request=request,
        )
        return response


class AsyncOfflineHTTPClient(httpx.AsyncClient):
    async def send(self, request, **kwargs):
        response = httpx.Response(
            status_code=200,
            json={"data": {}},
            request=request,
        )
        return response
